# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from cil.framework import AcquisitionGeometry
from cil.utilities.dataexample import SIMULATED_PARALLEL_BEAM_DATA, SIMULATED_CONE_BEAM_DATA, SIMULATED_SPHERE_VOLUME

import unittest
from scipy.fft  import fft, ifft
import numpy as np
from utils import has_tigre, has_gpu_tigre, has_ipp, has_astra, has_gpu_astra

if has_tigre:
    from cil.plugins.tigre import ProjectionOperator as ProjectionOperator
    from cil.plugins.tigre import FBP as FBP_tigre
    from tigre.utilities.filtering import ramp_flat, filter

    from cil.recon.Reconstructor import Reconstructor # checks on baseclass
    from cil.recon.FBP import GenericFilteredBackProjection # checks on baseclass
    from cil.recon import FDK, FBP

has_tigre_gpu = has_gpu_tigre()
if  not has_tigre_gpu:
    print("Unable to run TIGRE tests")

has_astra_gpu = has_gpu_astra()
if  not has_astra_gpu:
    print("Unable to run ASTRA tests")
    
class Test_Reconstructor(unittest.TestCase):

    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 255
        voxel_num_z = 15

        mag = 2
        src_to_obj = 50
        src_to_det = src_to_obj * mag

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 1000
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Cone2D([0,-src_to_obj],[0,src_to_det-src_to_obj])\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Cone3D([0,-src_to_obj,0],[0,src_to_det-src_to_obj,0])\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])        
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random')
        self.ig3D = self.ag3D.get_ImageGeometry()


    @unittest.skipUnless(has_tigre, "TIGRE not installed")
    def test_defaults(self):
        reconstructor = Reconstructor(self.ad3D)
        self.assertEqual(id(reconstructor.input),id(self.ad3D))
        self.assertEqual(reconstructor.image_geometry,self.ig3D)
        self.assertEqual(reconstructor.backend, 'tigre')


    @unittest.skipUnless(has_tigre, "TIGRE not installed")
    def test_set_input(self):
        reconstructor = Reconstructor(self.ad3D)
        self.assertEqual(id(reconstructor.input),id(self.ad3D))

        ag3D_new = self.ad3D.copy()
        reconstructor.set_input(ag3D_new)
        self.assertEqual(id(reconstructor.input),id(ag3D_new))

        ag3D_new = self.ad3D.get_slice(vertical='centre')
        with self.assertRaises(ValueError):
            reconstructor.set_input(ag3D_new)

        with self.assertRaises(TypeError):
            reconstructor = Reconstructor(self.ag3D)


    @unittest.skipUnless(has_tigre, "TIGRE not installed")
    def test_weak_input(self):

        data = self.ad3D.copy()
        reconstructor = Reconstructor(data)
        self.assertEqual(id(reconstructor.input),id(data))
        del data

        with self.assertRaises(ValueError):
            reconstructor.input

        reconstructor.set_input(self.ad3D)
        self.assertEqual(id(reconstructor.input),id(self.ad3D))


    @unittest.skipUnless(has_tigre, "TIGRE not installed")
    def test_set_image_data(self):
        reconstructor = Reconstructor(self.ad3D)

        self.ig3D.voxel_num_z = 1
        reconstructor.set_image_geometry(self.ig3D)
        self.assertEqual(reconstructor.image_geometry,self.ig3D)


    @unittest.skipUnless(has_tigre, "TIGRE not installed")
    def test_set_backend(self):
        
        with self.assertRaises(ValueError):
            reconstructor = Reconstructor(self.ad3D, backend='unsupported_backend')


class Test_GenericFilteredBackProjection(unittest.TestCase):

    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 16
        voxel_num_z = 4

        mag = 2
        src_to_obj = 50
        src_to_det = src_to_obj * mag

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 36
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Cone2D([0,-src_to_obj],[0,src_to_det-src_to_obj])\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Cone3D([0,-src_to_obj,0],[0,src_to_det-src_to_obj,0])\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])        
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random')
        self.ig3D = self.ag3D.get_ImageGeometry()
        

    @unittest.skipUnless(has_tigre, "TIGRE not installed")
    def check_defaults(self, reconstructor):
        self.assertEqual(reconstructor.filter, 'ram-lak')
        self.assertEqual(reconstructor.fft_order, 8)
        self.assertFalse(reconstructor.filter_inplace)
        self.assertIsNone(reconstructor._weights)

        filter = reconstructor.get_filter_array()
        self.assertEqual(type(filter), np.ndarray)
        self.assertEqual(len(filter), 2**8)
        self.assertEqual(filter[0], 0)
        self.assertEqual(filter[128],1.0)
        self.assertEqual(filter[1],filter[255])

        self.assertEqual(reconstructor.image_geometry,self.ig3D)


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_defaults(self):

        reconstructor = GenericFilteredBackProjection(self.ad3D)
        self.check_defaults(reconstructor)


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_reset(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)
        reconstructor.set_fft_order(10)

        arr = reconstructor.get_filter_array()
        arr.fill(0)
        reconstructor.set_filter(arr)

        ig = self.ig3D.copy()
        ig.num_voxels_x = 4
        reconstructor.set_image_geometry(ig)

        reconstructor.set_filter_inplace(True)

        reconstructor.reset()
        self.check_defaults(reconstructor)


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_set_filter(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)

        with self.assertRaises(ValueError):
            reconstructor.set_filter("unsupported_filter")

        filter = reconstructor.get_filter_array()
        filter_new =filter *0.5
        reconstructor.set_filter(filter_new)
        self.assertEqual(reconstructor.filter, 'custom')
        filter = reconstructor.get_filter_array()
        np.testing.assert_array_equal(filter,filter_new)

        with self.assertRaises(ValueError):
            reconstructor.set_filter(filter[1:-1])


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_set_fft_order(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)
        reconstructor.set_fft_order(10)
        self.assertEqual(reconstructor.fft_order, 10)

        with self.assertRaises(ValueError):
            reconstructor.set_fft_order(2)


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_set_filter_inplace(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)
        reconstructor.set_filter_inplace(True)
        self.assertTrue(reconstructor.filter_inplace)

        with self.assertRaises(TypeError):
            reconstructor.set_filter_inplace('unsupported_value')



class Test_FDK(unittest.TestCase):

    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 16
        voxel_num_z = 4

        mag = 2
        src_to_obj = 50
        src_to_det = src_to_obj * mag

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 36
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Cone2D([0,-src_to_obj],[0,src_to_det-src_to_obj])\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Cone3D([0,-src_to_obj,0],[0,src_to_det-src_to_obj,0])\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])        
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random')
        self.ig3D = self.ag3D.get_ImageGeometry()
        

    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_set_filter(self):
    
        reconstructor = FDK(self.ad3D)
        filter = reconstructor.get_filter_array()
        filter_new =filter *0.5
        reconstructor.set_filter(filter_new)

        reconstructor.set_fft_order(10)
        with self.assertRaises(ValueError):
            reconstructor._pre_filtering(self.ad3D)


    @unittest.skipUnless(has_tigre and has_ipp, "Prerequisites not met")
    def test_filtering(self):
        ag = AcquisitionGeometry.create_Cone3D([0,-1,0],[0,2,0])\
            .set_panel([64,3],[0.1,0.1])\
            .set_angles([0,90])

        ad = ag.allocate('random',seed=0)

        reconstructor = FDK(ad)
        out1 = ad.copy()
        reconstructor._pre_filtering(out1)

        #by hand
        filter = reconstructor.get_filter_array()
        reconstructor._calculate_weights(ag)
        pad0 = (len(filter)-ag.pixel_num_h)//2
        pad1 = len(filter)-ag.pixel_num_h-pad0

        out2 = ad.array.copy()
        out2*=reconstructor._weights
        for i in range(2):
            proj_padded = np.zeros((ag.pixel_num_v,len(filter)))
            proj_padded[:,pad0:-pad1] = out2[i]
            filtered_proj=fft(proj_padded,axis=-1)
            filtered_proj*=filter
            filtered_proj=ifft(filtered_proj,axis=-1)
            out2[i]=np.real(filtered_proj)[:,pad0:-pad1]

        diff = (out1-out2).abs().max()
        self.assertLess(diff, 1e-5)


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_weights(self):
        ag = AcquisitionGeometry.create_Cone3D([0,-1,0],[0,2,0])\
            .set_panel([3,4],[0.1,0.2])\
            .set_angles([0,90])
        ad = ag.allocate(0)

        reconstructor = FDK(ad)
        reconstructor._calculate_weights(ag)
        weights = reconstructor._weights

        scaling =  7.5 * np.pi
        weights_new = np.ones_like(weights)

        det_size_x = ag.pixel_size_h*ag.pixel_num_h
        det_size_y = ag.pixel_size_v*ag.pixel_num_v

        ray_length_z = 3
        for j in range(4):
            ray_length_y = -det_size_y/2 +  ag.pixel_size_v * (j+0.5)
            for i in range(3):
                ray_length_x = -det_size_x/2 +  ag.pixel_size_h * (i+0.5)
                ray_length = (ray_length_x**2+ray_length_y**2+ray_length_z**2)**0.5
                weights_new[j,i] = scaling*ray_length_z/ray_length

        diff = np.max(np.abs(weights - weights_new))
        self.assertLess(diff, 1e-5)


class Test_FBP(unittest.TestCase):
    
    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 16
        voxel_num_z = 4

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 36
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Parallel2D()\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Parallel3D()\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])        
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random')
        self.ig3D = self.ag3D.get_ImageGeometry()
        

    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_set_filter(self):
        reconstructor = FBP(self.ad3D)
        filter = reconstructor.get_filter_array()
        filter_new =filter *0.5
        reconstructor.set_filter(filter_new)

        reconstructor.set_fft_order(10)
        with self.assertRaises(ValueError):
            reconstructor._pre_filtering(self.ad3D)


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_split_processing(self):
        reconstructor = FBP(self.ad3D)
        
        self.assertEqual(reconstructor.slices_per_chunk, 0)

        reconstructor.set_split_processing(1)
        self.assertEqual(reconstructor.slices_per_chunk, 1)

        reconstructor.reset()
        self.assertEqual(reconstructor.slices_per_chunk, 0)


    @unittest.skipUnless(has_tigre and has_ipp, "Prerequisites not met")
    def test_filtering(self):
        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_panel([64,3],[0.1,0.1])\
            .set_angles([0,90])

        ad = ag.allocate('random',seed=0)

        reconstructor = FBP(ad)
        out1 = ad.copy()
        reconstructor._pre_filtering(out1)

        #by hand
        filter = reconstructor.get_filter_array()
        reconstructor._calculate_weights(ag)
        pad0 = (len(filter)-ag.pixel_num_h)//2
        pad1 = len(filter)-ag.pixel_num_h-pad0

        out2 = ad.array.copy()
        out2*=reconstructor._weights
        for i in range(2):
            proj_padded = np.zeros((ag.pixel_num_v,len(filter)))
            proj_padded[:,pad0:-pad1] = out2[i]
            filtered_proj=fft(proj_padded,axis=-1)
            filtered_proj*=filter
            filtered_proj=ifft(filtered_proj,axis=-1)
            out2[i]=np.real(filtered_proj)[:,pad0:-pad1]

        diff = (out1-out2).abs().max()
        self.assertLess(diff, 1e-5)


    @unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
    def test_weights(self):
        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_panel([3,4],[0.1,0.2])\
            .set_angles([0,90])
        ad = ag.allocate(0)

        reconstructor = FBP(ad)
        reconstructor._calculate_weights(ag)
        weights = reconstructor._weights

        scaling =  (2 * np.pi/ ag.num_projections) / ( 4 * ag.pixel_size_h ) 
        weights_new = np.ones_like(weights) * scaling

        np.testing.assert_allclose(weights,weights_new)


    @unittest.skipUnless(has_astra, "ASTRA not installed")
    def test_set_backend(self):

        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_panel([3,4],[0.1,0.2])\
            .set_angles([0,90])\
            .set_labels(['vertical','angle','horizontal'])
        ad = ag.allocate(0)

        reconstructor = FBP(ad, backend='astra')

        #Reconstructor backend raises an error if backend isn't supported
        with self.assertRaises(ValueError):
            reconstructor = FBP(ad, backend='unsupported_backend')

        #Reconstructor backend raises an error if dataorder isn't compatible
        with self.assertRaises(ValueError):
            reconstructor = FBP(ad, backend='tigre')


class Test_FDK_results(unittest.TestCase):
    
    def setUp(self):

        self.acq_data = SIMULATED_CONE_BEAM_DATA.get()
        self.img_data = SIMULATED_SPHERE_VOLUME.get()

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_3D(self):

        reconstructor = FDK(self.acq_data)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)    

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)     


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_2D(self):

        data2D = self.acq_data.get_slice(vertical='centre')
        img_data2D = self.img_data.get_slice(vertical='centre')

        reconstructor = FDK(data2D)
        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), img_data2D.as_array(),atol=1e-3)    

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)   


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_with_tigre(self):

        fbp_tigre = FBP_tigre(self.ig, self.ag)
        reco_tigre = fbp_tigre(self.acq_data)
    
        #fbp CIL with TIGRE's filter
        reconstructor_cil = FDK(self.acq_data)
        n = 2**reconstructor_cil.fft_order
        ramp = ramp_flat(n)
        filt = filter('ram_lak',ramp[0],n,1,False)

        reconstructor_cil = FDK(self.acq_data)
        reconstructor_cil.set_filter(filt)
        reco_cil = reconstructor_cil.run(verbose=0)

        #with the same filter results should be virtually identical
        np.testing.assert_allclose(reco_cil.as_array(), reco_tigre.as_array(),atol=1e-8) 


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_inplace_filtering(self):

        reconstructor = FDK(self.acq_data)
        reco = reconstructor.run(verbose=0)

        data_filtered= self.acq_data.copy()
        reconstructor_inplace = FDK(data_filtered)
        reconstructor_inplace.set_filter_inplace(True)
        reconstructor_inplace.run(out=reco, verbose=0)

        diff = (data_filtered - self.acq_data).abs().mean()
        self.assertGreater(diff,0.8)

          
class Test_FBP_results(unittest.TestCase):
    
    def setUp(self):

        self.acq_data = SIMULATED_PARALLEL_BEAM_DATA.get()
        self.img_data = SIMULATED_SPHERE_VOLUME.get()

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_3D_tigre(self):

        reconstructor = FBP(self.acq_data)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)    

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)  


    @unittest.skipUnless(has_astra and has_astra_gpu and has_ipp, "ASTRA or IPP not installed")
    def test_results_3D_astra(self):

        self.acq_data.reorder('astra')
        reconstructor = FBP(self.acq_data, backend='astra')

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)    

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)  


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_3D_split(self):

        reconstructor = FBP(self.acq_data)
        reconstructor.set_split_processing(1)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)    

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)   


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_2D_tigre(self):

        data2D = self.acq_data.get_slice(vertical='centre')
        img_data2D = self.img_data.get_slice(vertical='centre')

        reconstructor = FBP(data2D)
        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), img_data2D.as_array(),atol=1e-3)    

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)    


    @unittest.skipUnless(has_astra and has_astra_gpu and has_ipp, "ASTRA or IPP not installed")
    def test_results_2D_astra(self):

        data2D = self.acq_data.get_slice(vertical='centre')
        data2D.reorder('astra')
        img_data2D = self.img_data.get_slice(vertical='centre')

        reconstructor = FBP(data2D, backend='astra')
        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), img_data2D.as_array(),atol=1e-3)    

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)    


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_with_tigre(self):

        fbp_tigre = FBP_tigre(self.ig, self.ag)
        reco_tigre = fbp_tigre(self.acq_data)
    
        #fbp CIL with TIGRE's filter
        reconstructor_cil = FBP(self.acq_data)
        n = 2**reconstructor_cil.fft_order
        ramp = ramp_flat(n)
        filt = filter('ram_lak',ramp[0],n,1,False)

        reconstructor_cil = FBP(self.acq_data)
        reconstructor_cil.set_filter(filt)
        reco_cil = reconstructor_cil.run(verbose=0)

        #with the same filter results should be virtually identical
        np.testing.assert_allclose(reco_cil.as_array(), reco_tigre.as_array(),atol=1e-8) 


    @unittest.skipUnless(has_tigre and has_tigre_gpu and has_ipp, "TIGRE or IPP not installed")
    def test_results_inplace_filtering(self):

        reconstructor = FBP(self.acq_data)
        reco = reconstructor.run(verbose=0)

        data_filtered= self.acq_data.copy()
        reconstructor_inplace = FBP(data_filtered)
        reconstructor_inplace.set_filter_inplace(True)
        reconstructor_inplace.run(out=reco, verbose=0)

        diff = (data_filtered - self.acq_data).abs().mean()
        self.assertGreater(diff,0.8)


# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

try:
    from ccpi.filters import regularisers
    from ccpi.filters.cpu_regularisers import TV_ENERGY
except ImportError as ie:
    raise ImportError(ie , "\n\n", 
                      "This plugin requires the additional package ccpi-regularisation\n" +
                      "Please install it via conda as ccpi-regulariser from the ccpi channel\n"+
                      "Minimal version is 20.04")


from cil.framework import DataOrder
from cil.optimisation.functions import Function
import numpy as np
import warnings
from numbers import Number

class RegulariserFunction(Function):
    def proximal(self, x, tau, out=None):
        '''Generic proximal method for a RegulariserFunction
        
        :param x: image to be regularised
        :type x: an ImageData
        :param tau: 
        :type tau: Number
        :param out: a placeholder for the result
        :type out: same as x: ImageData
        
        If the ImageData contains complex data, rather than the default float32, the regularisation
        is run indipendently on the real and imaginary part.
        '''

        self.check_input(x)
        arr = x.as_array()
        if arr.dtype in [np.complex, np.complex64]:
            # do real and imag part indep
            in_arr = np.asarray(arr.real, dtype=np.float32, order='C')
            res, info = self.proximal_numpy(in_arr, tau)
            arr.real = res[:]
            in_arr = np.asarray(arr.imag, dtype=np.float32, order='C')
            res, info = self.proximal_numpy(in_arr, tau)
            arr.imag = res[:]
            self.info = info
            if out is not None:
                out.fill(arr)
            else:
                out = x.copy()
                out.fill(arr)
                return out
        else:
            arr = np.asarray(x.as_array(), dtype=np.float32, order='C')
            res, info = self.proximal_numpy(arr, tau)
            self.info = info
            if out is not None:
                out.fill(res)
            else:
                out = x.copy()
                out.fill(res)
                return out
    def proximal_numpy(self, xarr, tau):
        raise NotImplementedError('Please implement proximal_numpy')

    def check_input(self, input):
        pass

class TV_Base(RegulariserFunction):
    def __call__(self,x):
        in_arr = np.asarray(x.as_array(), dtype=np.float32, order='C')
        EnergyValTV = TV_ENERGY(in_arr, in_arr, self.alpha, 2)
        return 0.5*EnergyValTV[0]

    def convex_conjugate(self,x):     
        return 0.0


class FGP_TV(TV_Base):
    def __init__(self, alpha=1, max_iteration=100, tolerance=0, isotropic=True, nonnegativity=True, device='cpu'):
        '''Creator of FGP_TV Function


        :param alpha: regularisation parameter
        :type alpha: number, default 1
        :param isotropic: Whether it uses L2 (isotropic) or L1 (unisotropic) norm
        :type isotropic: boolean, default True, can range between 1 and 2
        :param nonnegativity: Whether to add the non-negativity constraint
        :type nonnegativity: boolean, default True
        :param max_iteration: max number of sub iterations. The algorithm will iterate up to this number of iteration or up to when the tolerance has been reached
        :type max_iteration: integer, default 100
        :param tolerance: minimum difference between previous iteration of the algorithm that determines the stop of the iteration earlier than max_iteration. If set to 0 only the max_iteration will be used as stop criterion.
        :type tolerance: float, default 0
        :param device: determines if the code runs on CPU or GPU
        :type device: string, default 'cpu', can be 'gpu' if GPU is installed
        '''
        if isotropic == True:
            self.methodTV = 0
        else:
            self.methodTV = 1

        if nonnegativity == True:
            self.nonnegativity = 1
        else:
            self.nonnegativity = 0

        self.alpha = alpha
        self.max_iteration = max_iteration
        self.tolerance = tolerance
        self.nonnegativity = nonnegativity
        self.device = device # string for 'cpu' or 'gpu'

    def proximal_numpy(self, in_arr, tau):
        res , info = regularisers.FGP_TV(\
              in_arr,\
              self.alpha * tau,\
              self.max_iteration,\
              self.tolerance,\
              self.methodTV,\
              self.nonnegativity,\
              self.device)
        return res, info
    
    def __rmul__(self, scalar):
        '''Define the multiplication with a scalar
        
        this changes the regularisation parameter in the plugin'''
        if not isinstance (scalar, Number):
            raise NotImplemented
        else:
            self.alpha *= scalar
            return self
    def check_input(self, input):
        if input.geometry.length > 3:
            raise ValueError('{} cannot work on more than 3D. Got {}'.format(self.__class__.__name__, input.geometry.length))
        
class TGV(RegulariserFunction):

    def __init__(self, alpha=1, gamma=1, max_iteration=100, tolerance=0, device='cpu' , **kwargs):
        '''Creator of Total Generalised Variation Function 

        :param alpha: regularisation parameter
        :type alpha: number, default 1
        :param gamma: ratio of TGV terms
        :type gamma: number, default 1, can range between 1 and 2
        :param max_iteration: max number of sub iterations. The algorithm will iterate up to this number of iteration or up to when the tolerance has been reached
        :type max_iteration: integer, default 100
        :param tolerance: minimum difference between previous iteration of the algorithm that determines the stop of the iteration earlier than max_iteration. If set to 0 only the max_iteration will be used as stop criterion.
        :type tolerance: float, default 0
        :param device: determines if the code runs on CPU or GPU
        :type device: string, default 'cpu', can be 'gpu' if GPU is installed
        
        '''
        
        self.alpha = alpha
        self.gamma = gamma
        self.max_iteration = max_iteration
        self.tolerance = tolerance
        self.device = device

        if kwargs.get('iter_TGV', None) is not None:
            # raise ValueError('iter_TGV parameter has been superseded by num_iter. Use that instead.')
            self.num_iter = kwargs.get('iter_TGV')
        
    def __call__(self,x):
        warnings.warn("{}: the __call__ method is not implemented. Returning NaN.".format(self.__class__.__name__))
        return np.nan
    @property
    def gamma(self):
        return self.__gamma
    @gamma.setter
    def gamma(self, value):
        if value <= 2 and value >= 1:
            self.__gamma = value
    @property
    def alpha2(self):
        return self.alpha1 * self.gamma
    @property
    def alpha1(self):
        return 1.
    
    def proximal_numpy(self, in_arr, tau):
        res , info = regularisers.TGV(in_arr,
              self.alpha * tau,
              self.alpha1,
              self.alpha2,
              self.max_iteration,
              self.LipshitzConstant,
              self.tolerance,
              self.device)
                
        # info: return number of iteration and reached tolerance
        # https://github.com/vais-ral/CCPi-Regularisation-Toolkit/blob/master/src/Core/regularisers_CPU/TGV_core.c#L168
        # Stopping Criteria  || u^k - u^(k-1) ||_{2} / || u^{k} ||_{2}    
        return res, info
    
    def convex_conjugate(self, x):
        warnings.warn("{}: the convex_conjugate method is not implemented. Returning NaN.".format(self.__class__.__name__))
        return np.nan
        
    def __rmul__(self, scalar):
        '''Define the multiplication with a scalar
        
        this changes the regularisation parameter in the plugin'''
        if not isinstance (scalar, Number):
            raise NotImplemented
        else:
            self.alpha *= scalar
            return self

        # f = TGV()
        # f = alpha * f

    def check_input(self, input):
        if len(input.dimension_labels) == 2:
            self.LipshitzConstant = 12
        elif len(input.dimension_labels) == 3:
            self.LipshitzConstant = 16 # Vaggelis to confirm
        else:
            raise ValueError('{} cannot work on more than 3D. Got {}'.format(self.__class__.__name__, input.geometry.length))
        

class FGP_dTV(RegulariserFunction):
    '''Creator of FGP_dTV Function

        :param reference: reference image
        :type reference: ImageData
        :param alpha: regularisation parameter
        :type alpha: number, default 1
        :param max_iteration: max number of sub iterations. The algorithm will iterate up to this number of iteration or up to when the tolerance has been reached
        :type max_iteration: integer, default 100
        :param tolerance: minimum difference between previous iteration of the algorithm that determines the stop of the iteration earlier than max_iteration. If set to 0 only the max_iteration will be used as stop criterion.
        :type tolerance: float, default 0
        :param eta: smoothing constant to calculate gradient of the reference
        :type eta: number, default 0.01
        :param isotropic: Whether it uses L2 (isotropic) or L1 (anisotropic) norm
        :type isotropic: boolean, default True, can range between 1 and 2
        :param nonnegativity: Whether to add the non-negativity constraint
        :type nonnegativity: boolean, default True
        :param device: determines if the code runs on CPU or GPU
        :type device: string, default 'cpu', can be 'gpu' if GPU is installed
        '''
    def __init__(self, reference, alpha=1, max_iteration=100,
                 tolerance=0, eta=0.01, isotropic=True, nonnegativity=True, device='cpu'):

        if isotropic == True:
            self.methodTV = 0
        else:
            self.methodTV = 1

        if nonnegativity == True:
            self.nonnegativity = 1
        else:
            self.nonnegativity = 0

        self.alpha = alpha
        self.max_iteration = max_iteration
        self.tolerance = tolerance
        self.device = device # string for 'cpu' or 'gpu'
        self.reference = np.asarray(reference.as_array(), dtype=np.float32)
        self.eta = eta
        
    def __call__(self,x):
        warnings.warn("{}: the __call__ method is not implemented. Returning NaN.".format(self.__class__.__name__))
        return np.nan

    def proximal_numpy(self, in_arr, tau):
        res , info = regularisers.FGP_dTV(\
                in_arr,\
                self.reference,\
                self.alpha * tau,\
                self.max_iteration,\
                self.tolerance,\
                self.eta,\
                self.methodTV,\
                self.nonnegativity,\
                self.device)
        return res, info

    def convex_conjugate(self, x):
        warnings.warn("{}: the convex_conjugate method is not implemented. Returning NaN.".format(self.__class__.__name__))
        return np.nan
    
    def __rmul__(self, scalar):
        '''Define the multiplication with a scalar
        
        this changes the regularisation parameter in the plugin'''
        if not isinstance (scalar, Number):
            raise NotImplemented
        else:
            self.alpha *= scalar
            return self

    def check_input(self, input):
        if input.geometry.length > 3:
            raise ValueError('{} cannot work on more than 3D. Got {}'.format(self.__class__.__name__, input.geometry.length))

class TNV(RegulariserFunction):
    
    def __init__(self,alpha=1, max_iteration=100, tolerance=0):
        '''Creator of TNV Function

        :param alpha: regularisation parameter
        :type alpha: number, default 1
        :param max_iteration: max number of sub iterations. The algorithm will iterate up to this number of iteration or up to when the tolerance has been reached
        :type max_iteration: integer, default 100
        :param tolerance: minimum difference between previous iteration of the algorithm that determines the stop of the iteration earlier than max_iteration. If set to 0 only the max_iteration will be used as stop criterion.
        :type tolerance: float, default 0
        '''
        # set parameters
        self.alpha = alpha
        self.max_iteration = max_iteration
        self.tolerance = tolerance
        
    def __call__(self,x):
        warnings.warn("{}: the __call__ method is not implemented. Returning NaN.".format(self.__class__.__name__))
        return np.nan
    
    def proximal_numpy(self, in_arr, tau):
        if in_arr.ndim != 3:
            # https://github.com/vais-ral/CCPi-Regularisation-Toolkit/blob/413c6001003c6f1272aeb43152654baaf0c8a423/src/Python/src/cpu_regularisers.pyx#L584-L588
            raise ValueError('Only 3D data is supported. Passed data has {} dimensions'.format(in_arr.ndim))
        res = regularisers.TNV(in_arr, 
              self.alpha * tau,
              self.max_iteration,
              self.tolerance)
        return res, []

    def convex_conjugate(self, x):
        warnings.warn("{}: the convex_conjugate method is not implemented. Returning NaN.".format(self.__class__.__name__))
        return np.nan

    def __rmul__(self, scalar):
        '''Define the multiplication with a scalar
        
        this changes the regularisation parameter in the plugin'''
        if not isinstance (scalar, Number):
            raise NotImplemented
        else:
            self.alpha *= scalar
            return self

    def check_input(self, input):
        '''TNV requires 2D+channel data with the first dimension as the channel dimension'''
        DataOrder.check_order_for_engine('cil', input.geometry)
        if ( input.geometry.channels == 1 ) or ( not input.geometry.length == 3) :
            raise ValueError('TNV requires 2D+channel data. Got {}'.format(input.geometry.dimension_labels))
        


import numpy as np
from cil.optimisation.operators import LinearOperator
from cil.utilities import dataexample
from cil.framework import AcquisitionGeometry
from cil.framework import DataOrder

class SimData(object):

    def _get_roi_3D(self):
        roi = [20,30,40]
        center_offset = [50,-20,5]

        self.ig_roi = self.ig.copy()
        self.ig_roi.voxel_num_x = roi[2]
        self.ig_roi.voxel_num_y = roi[1]
        self.ig_roi.voxel_num_z = roi[0]
        self.ig_roi.center_x = center_offset[2]*self.ig_roi.voxel_size_x
        self.ig_roi.center_y = center_offset[1]*self.ig_roi.voxel_size_y
        self.ig_roi.center_z = center_offset[0]*self.ig_roi.voxel_size_z

        index_roi = [None]*3
        for i in range(3):
            ind0 = center_offset[i] + (self.img_data.shape[i] - roi[i])//2
            ind1 = ind0 + roi[i]
            index_roi[i] = (ind0, ind1)

        self.gold_roi = self.img_data.array[index_roi[0][0]:index_roi[0][1],index_roi[1][0]:index_roi[1][1],index_roi[2][0]:index_roi[2][1]]


    def _get_roi_2D(self):
        roi = [30,40]
        center_offset = [-20,5]

        self.ig_roi = self.ig.copy()
        self.ig_roi.voxel_num_x = roi[1]
        self.ig_roi.voxel_num_y = roi[0]
        self.ig_roi.center_x = center_offset[1]*self.ig_roi.voxel_size_x
        self.ig_roi.center_y = center_offset[0]*self.ig_roi.voxel_size_y

        index_roi = [None]*2
        for i in range(2):
            ind0 = center_offset[i] + (self.img_data.shape[i] - roi[i])//2
            ind1 = ind0 + roi[i]
            index_roi[i] = (ind0, ind1)

        self.gold_roi = self.img_data.array[index_roi[0][0]:index_roi[0][1],index_roi[1][0]:index_roi[1][1]]


    def Cone3D(self):
        self.acq_data = dataexample.SIMULATED_CONE_BEAM_DATA.get()
        self.acq_data.reorder(self.backend)
        
        self.img_data = dataexample.SIMULATED_SPHERE_VOLUME.get()

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry

        self._get_roi_3D()


    def Parallel3D(self):
        self.acq_data = dataexample.SIMULATED_PARALLEL_BEAM_DATA.get()
        self.acq_data.reorder(self.backend)
        
        self.img_data = dataexample.SIMULATED_SPHERE_VOLUME.get()

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry

        self._get_roi_3D()


    def Cone2D(self):

        self.acq_data = dataexample.SIMULATED_CONE_BEAM_DATA.get().get_slice(vertical='centre')
        self.acq_data.reorder(self.backend)
        
        self.img_data = dataexample.SIMULATED_SPHERE_VOLUME.get().get_slice(vertical='centre')

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry

        self._get_roi_2D()


    def Parallel2D(self):

        self.acq_data = dataexample.SIMULATED_PARALLEL_BEAM_DATA.get().get_slice(vertical='centre')
        self.acq_data.reorder(self.backend)
        
        self.img_data = dataexample.SIMULATED_SPHERE_VOLUME.get().get_slice(vertical='centre')

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry

        self._get_roi_2D()


class TestCommon_ProjectionOperator_TOY(object):
    '''
    Tests behaviour of the operators on a variety of geometries.
    '''
    def Cone3D(self):
        '''
            These are all single cone beam projection geometries. Pixels of  1, 2, 0.5, 0.5, Voxels of 1, 2, 0.5, 0.25
        '''

        self.test_geometries=[]
        ag_test_1 = AcquisitionGeometry.create_Cone3D(source_position=[0,-1000,0],detector_position=[0,0,0])\
                                            .set_panel([16,16],[1,1])\
                                            .set_angles([0])
        ag_test_1.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_1))

        
        ig_test_1 = ag_test_1.get_ImageGeometry()
        norm_1 = 4
        self.test_geometries.append((ag_test_1, ig_test_1, 4))


        ag_test_2 = AcquisitionGeometry.create_Cone3D(source_position=[0,-1000,0],detector_position=[0,0,0])\
                                            .set_panel([16,16],[2,2])\
                                            .set_angles([0])
        ag_test_2.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_2))

        ig_test_2 = ag_test_2.get_ImageGeometry()
        norm_2 = 8
        self.test_geometries.append((ag_test_2, ig_test_2, norm_2))


        ag_test_3 = AcquisitionGeometry.create_Cone3D(source_position=[0,-1000,0],detector_position=[0,0,0])\
                                            .set_panel([16,16],[0.5,0.5])\
                                            .set_angles([0])
        ag_test_3.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_3))
        ig_test_3 = ag_test_3.get_ImageGeometry()

        norm_3 = 2
        self.test_geometries.append((ag_test_3, ig_test_3, norm_3))


        ag_test_4 = AcquisitionGeometry.create_Cone3D(source_position=[0,-1000,0],detector_position=[0,1000,0])\
                                            .set_panel([16,16],[0.5,0.5])\
                                            .set_angles([0])
        ag_test_4.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_4))
        ig_test_4 = ag_test_4.get_ImageGeometry()

        norm_4 = 1
        self.test_geometries.append((ag_test_4, ig_test_4, norm_4))


    def Cone2D(self):
        '''
            These are all single cone beam projection geometries. Pixels of  1, 2, 0.5, 0.5, Voxels of 1, 2, 0.5, 0.25
        '''

        self.test_geometries=[]
        ag_test_1 = AcquisitionGeometry.create_Cone2D(source_position=[0,-1000],detector_position=[0,0])\
                                            .set_panel(16,1)\
                                            .set_angles([0])
        ag_test_1.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_1))

        
        ig_test_1 = ag_test_1.get_ImageGeometry()
        norm_1 = 4
        self.test_geometries.append((ag_test_1, ig_test_1, 4))


        ag_test_2 = AcquisitionGeometry.create_Cone2D(source_position=[0,-1000],detector_position=[0,0])\
                                            .set_panel(16,2)\
                                            .set_angles([0])
        ag_test_2.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_2))

        ig_test_2 = ag_test_2.get_ImageGeometry()
        norm_2 = 8
        self.test_geometries.append((ag_test_2, ig_test_2, norm_2))


        ag_test_3 = AcquisitionGeometry.create_Cone2D(source_position=[0,-1000],detector_position=[0,0])\
                                            .set_panel(16,0.5)\
                                            .set_angles([0])
        ag_test_3.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_3))
        ig_test_3 = ag_test_3.get_ImageGeometry()

        norm_3 = 2
        self.test_geometries.append((ag_test_3, ig_test_3, norm_3))


        ag_test_4 = AcquisitionGeometry.create_Cone2D(source_position=[0,-1000],detector_position=[0,1000])\
                                            .set_panel(16,0.5)\
                                            .set_angles([0])
        ag_test_4.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_4))
        ig_test_4 = ag_test_4.get_ImageGeometry()

        norm_4 = 1
        self.test_geometries.append((ag_test_4, ig_test_4, norm_4))


    def Parallel3D(self):
        '''
            These are all single parallel beam projection geometries. Pixels & voxels of  1, 2, 0.5
        '''

        self.test_geometries=[]
        ag_test_1 = AcquisitionGeometry.create_Parallel3D()\
                                            .set_panel([16,16],[1,1])\
                                            .set_angles([0])
        ag_test_1.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_1))

        
        ig_test_1 = ag_test_1.get_ImageGeometry()
        norm_1 = 4
        self.test_geometries.append((ag_test_1, ig_test_1, norm_1))

        ag_test_2 = AcquisitionGeometry.create_Parallel3D()\
                                            .set_panel([16,16],[2,2])\
                                            .set_angles([0])
        ag_test_2.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_2))


        ig_test_2 = ag_test_2.get_ImageGeometry()
        norm_2 = 8
        self.test_geometries.append((ag_test_2, ig_test_2, norm_2))


        ag_test_3 = AcquisitionGeometry.create_Parallel3D()\
                                            .set_panel([16,16],[0.5,0.5])\
                                            .set_angles([0])
        ag_test_3.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_3))


        ig_test_3 = ag_test_3.get_ImageGeometry()
        norm_3 = 2
        self.test_geometries.append((ag_test_3, ig_test_3, norm_3))


    def Parallel2D(self):
        '''
            These are all single parallel beam projection geometries. Pixels & voxels of  1, 2, 0.5
        '''

        self.test_geometries=[]
        ag_test_1 = AcquisitionGeometry.create_Parallel2D()\
                                            .set_panel(16,1)\
                                            .set_angles([0])

        ag_test_1.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_1))

        ig_test_1 = ag_test_1.get_ImageGeometry()
        norm_1 = 4
        self.test_geometries.append((ag_test_1, ig_test_1, norm_1))

        ag_test_2 = AcquisitionGeometry.create_Parallel2D()\
                                            .set_panel(16,2)\
                                            .set_angles([0])
        ag_test_2.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_2))


        ig_test_2 = ag_test_2.get_ImageGeometry()
        norm_2 = 8
        self.test_geometries.append((ag_test_2, ig_test_2, norm_2))


        ag_test_3 = AcquisitionGeometry.create_Parallel2D()\
                                            .set_panel(16,0.5)\
                                            .set_angles([0])
        ag_test_3.set_labels(DataOrder.get_order_for_engine(self.backend, ag_test_3))

        ig_test_3 = ag_test_3.get_ImageGeometry()
        norm_3 = 2
        self.test_geometries.append((ag_test_3, ig_test_3, norm_3))


    def test_norm(self):
        count = 1
        for ag, ig, norm in self.test_geometries:
            Op = self.ProjectionOperator(ig, ag,  **self.PO_args)
            n = Op.norm()

            diff =abs(n-norm)
            self.assertLess(diff, self.tolerance_norm, "Geometry: {0} Calculated norm {1}, expected {2}".format(count, n, norm))


    def test_linearity(self):
        count = 1
        for ag, ig, _ in self.test_geometries:
            Op = self.ProjectionOperator(ig, ag,  **self.PO_args)
            res = LinearOperator.dot_test(Op, tolerance=self.tolerance_linearity)
            self.assertTrue(res, "Geometry: {0} Dot-test out of tolerance {1}".format(count, self.tolerance_linearity))
            count +=1



class TestCommon_ProjectionOperator(object):

    '''
    These are all single projection geometries approximating parallel beam.
    '''

    def Cone3D(self):
        self.ag = AcquisitionGeometry.create_Cone3D(source_position=[0,-100000,0],detector_position=[0,0,0])\
                                            .set_panel([16,16],[1,1])\
                                            .set_angles([0])\
                                            .set_labels(['vertical','horizontal'])

    def Cone2D(self):
        self.ag = AcquisitionGeometry.create_Cone2D(source_position=[0,-100000],detector_position=[0,0])\
                                            .set_panel(16,1)\
                                            .set_angles([0])\
                                            .set_labels(['horizontal'])

    def Parallel3D(self):
        self.ag = AcquisitionGeometry.create_Parallel3D()\
                                            .set_panel([16,16],[1,1])\
                                            .set_angles([0])\
                                            .set_labels(['vertical','horizontal'])

    def Parallel2D(self):
        self.ag = AcquisitionGeometry.create_Parallel2D()\
                                            .set_panel(16,1)\
                                            .set_angles([0])\
                                            .set_labels(['horizontal'])

    def test_forward_projector(self):

        #create checker-board image
        res = np.zeros((16,16,16))
        ones = np.ones((4,16,4))
        for k in range(4):
            for i in range(4):
                if (i + k)% 2 == 0: 
                    res[k*4:(k+1)*4,:,i*4:(i+1)*4] = ones

        #create checker-board forward projection for parallel rays
        checker = np.zeros((16,16))
        ones = np.ones((4,4))
        for j in range(4):
            for i in range(4):
                if (i + j)% 2 == 0: 
                    checker[j*4:(j+1)*4,i*4:(i+1)*4] = ones * 16

        if self.ag.dimension == '2D':
            checker = checker[0]
            res = res[0]

        ig = self.ag.get_ImageGeometry()
        volume = ig.allocate(0)

        volume.fill(res)

        Op = self.ProjectionOperator(ig, self.ag, **self.PO_args)
        fp = Op.direct(volume)

        np.testing.assert_allclose(fp.array, checker, atol = self.tolerance_fp)


    def test_backward_projector(self):

        #create checker-board projection
        checker = np.zeros((16,16))
        ones = np.ones((4,4))
        for j in range(4):
            for i in range(4):
                if (i + j)% 2 == 0: 
                    checker[j*4:(j+1)*4,i*4:(i+1)*4] = ones

        #create backprojection of checker-board
        res = np.zeros((16,16,16))
        ones = np.ones((4,16,4))
        for k in range(4):
            for i in range(4):
                if (i + k)% 2 == 0: 
                    res[k*4:(k+1)*4,:,i*4:(i+1)*4] = ones

        if self.ag.dimension == '2D':
            checker = checker[0]
            res = res[0]

        ig = self.ag.get_ImageGeometry()
        data = self.ag.allocate(0)

        data.fill(checker)

        Op = self.ProjectionOperator(ig, self.ag, **self.PO_args)
        bp = Op.adjoint(data)

        if self.ag.geom_type == 'cone':
            #as cone beam res is not perfect grid
            np.testing.assert_allclose(bp.array, res, atol=1e-3)
        else:
            np.testing.assert_equal(bp.array, res)


class TestCommon_ProjectionOperator_SIM(SimData):
    '''
    Tests forward and backward operators function with and without 'out'
    '''
    def test_forward_projector(self):
        Op = self.ProjectionOperator(self.ig, self.ag, **self.PO_args)
        fp = Op.direct(self.img_data)
        np.testing.assert_allclose(fp.as_array(), self.acq_data.as_array(),atol=self.tolerance_fp)        

        fp2 = fp.copy()
        fp2.fill(0)
        Op.direct(self.img_data,out=fp2)
        np.testing.assert_allclose(fp.as_array(), fp2.as_array(),1e-8)    


    def test_backward_projectors_functionality(self):
        #this checks mechanics but not value
        Op = self.ProjectionOperator(self.ig, self.ag, **self.PO_args)
        bp = Op.adjoint(self.acq_data)

        bp2 = bp.copy()
        bp2.fill(0)
        Op.adjoint(self.acq_data,out=bp2)
        np.testing.assert_allclose(bp.as_array(), bp2.as_array(), 1e-8)    


class TestCommon_FBP_SIM(SimData):
    '''
    FBP tests on simulated data
    '''
    def test_FBP(self):

        self.FBP = self.FBP(self.ig, self.ag, **self.FBP_args)
        reco = self.FBP(self.acq_data)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=self.tolerance_fbp)    

        reco2 = reco.copy()
        reco2.fill(0)
        self.FBP(self.acq_data,out=reco2)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(),atol=1e-8)   


    def test_FBP_roi(self):
        self.FBP = self.FBP(self.ig_roi, self.ag, **self.FBP_args)
        reco = self.FBP(self.acq_data)
        np.testing.assert_allclose(reco.as_array(), self.gold_roi, atol=self.tolerance_fbp_roi)    



# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from cil.framework import DataProcessor, AcquisitionData, ImageData
import numpy
from numbers import Number


class Padder(DataProcessor):
    r'''
    Processor to pad an array, wrapping numpy.pad
    See https://numpy.org/doc/stable/reference/generated/numpy.pad.html
    '''

    @staticmethod
    def constant(pad_width=None, constant_values=0):
        '''
        Padder processor wrapping numpy.pad with mode `constant` 
        Pads with a constant value.

        Parameters
        ----------
        pad_width: int, tuple, dict
            The number of values padded to the edges of each axis
        constant_values: float, tuple, dict
            The values to set the padded values for each axis

        Notes
        -----
        If passed a single value it will pad symmetrically in all dimensions.
        If passed a tuple it will apply asymmetric padding in all dimensions. (before, after)
        If passed a dictionary it will apply the specified padding to the required dimension label: e.g.
        {'horizontal':(8, 23), 'vertical': 10}
        '''
        processor = Padder(pad_width=pad_width, mode='constant', constant_values=constant_values)
        return processor

    @staticmethod
    def edge(pad_width=None):
        '''Padder processor wrapping numpy.pad with mode `edge` 
        Pads with the edge values of array.

        Parameters
        ----------
        pad_width: int, tuple, dict
            The number of values padded to the edges of each axis

        Notes
        -----
        If passed a single value it will pad symmetrically in all dimensions.
        If passed a tuple it will apply asymmetric padding in all dimensions. (before, after)
        If passed a dictionary it will apply the specified padding to the required dimension label: e.g.
        {'horizontal':(8, 23), 'vertical': 10}
        '''

        processor = Padder(pad_width=pad_width, mode='edge')
        return processor
    
    @staticmethod
    def linear_ramp(pad_width=None, end_values=0):
        '''Padder processor wrapping numpy.pad with mode `linear_ramp` 
        Pads with the linear ramp between end_value and the array edge value.

        Parameters
        ----------
        pad_width: int, tuple, dict
            The number of values padded to the edges of each axis
        end_values: float, tuple, dict
            The values used for the ending value of the linear_ramp
 
        Notes
        -----
        If passed a single value it will pad symmetrically in all dimensions.
        If passed a tuple it will apply asymmetric padding in all dimensions. (before, after)
        If passed a dictionary it will apply the specified padding to the required dimension label: e.g.
        {'horizontal':(8, 23), 'vertical': 10}
        '''
        processor = Padder(pad_width=pad_width, mode='linear_ramp', end_values=end_values)
        return processor
    
    @staticmethod
    def reflect(pad_width=None):
        '''Padder processor wrapping numpy.pad with mode `reflect` 
        Pads with the reflection of the vector mirrored on the first and last values of the vector along each axis.
        
        Parameters
        ----------
        pad_width: int, tuple, dict
            The number of values padded to the edges of each axis

        Notes
        -----
        If passed a single value it will pad symmetrically in all dimensions.
        If passed a tuple it will apply asymmetric padding in all dimensions. (before, after)
        If passed a dictionary it will apply the specified padding to the required dimension label: e.g.
        {'horizontal':(8, 23), 'vertical': 10}
        '''
        processor = Padder(pad_width=pad_width, mode='reflect')
        return processor
    
    @staticmethod
    def symmetric(pad_width=None):
        r'''Padder processor wrapping numpy.pad with mode `symmetric`
        Pads with the reflection of the vector mirrored along the edge of the array.

        Parameters
        ----------
        pad_width: int, tuple, dict
            The number of values padded to the edges of each axis

        Notes
        -----
        If passed a single value it will pad symmetrically in all dimensions.
        If passed a tuple it will apply asymmetric padding in all dimensions. (before, after)
        If passed a dictionary it will apply the specified padding to the required dimension label: e.g.
        {'horizontal':(8, 23), 'vertical': 10}
        '''
        processor = Padder(pad_width=pad_width, mode='symmetric')
        return processor
    
    @staticmethod
    def wrap(pad_width=None):
        '''Padder processor wrapping numpy.pad with mode `wrap`
        Pads with the wrap of the vector along the axis. The first values are used to pad the end and the end values are used to pad the beginning.

        Parameters
        ----------
        pad_width: int, tuple, dict
            The number of values padded to the edges of each axis

        Notes
        -----
        If passed a single value it will pad symmetrically in all dimensions.
        If passed a tuple it will apply asymmetric padding in all dimensions. (before, after)
        If passed a dictionary it will apply the specified padding to the required dimension label: e.g.
        {'horizontal':(8, 23), 'vertical': 10}
        '''
        processor = Padder(pad_width=pad_width, mode='wrap')
        return processor

    def __init__(self,
                 mode='constant',
                 pad_width=None,
                 constant_values=0,
                 end_values=0):

        kwargs = {'mode': mode,
                'pad_width': pad_width,
                'constant_values': constant_values,
                'end_values': end_values}

        super(Padder, self).__init__(**kwargs)


    def check_input(self, data):

        if not ((isinstance(data, ImageData)) or 
                (isinstance(data, AcquisitionData))):
            raise TypeError('Processor supports only following data types:\n' +
                            ' - ImageData\n - AcquisitionData')

        elif (data.geometry == None):
            raise ValueError('Geometry is not defined.')

        elif self.mode not in ['constant', 'edge', 'linear_ramp', 'reflect', 'symmetric', 'wrap']:
            raise Exception("Wrong mode. One of the following is expected:\n" +
                            "constant, edge, linear_ramp, reflect, symmetric, wrap")

        elif (self.pad_width == None):
            raise ValueError('Please, specify pad_width')

        else:
            return True 


    def process(self, out=None):

        data = self.get_input()
        ndim = data.number_of_dimensions
        dimension_labels = data.dimension_labels
        geometry_0 = data.geometry
        # create a new geometry for the new dataset
        geometry = geometry_0.copy()

        pad_width_param = self._parse_param(data, self.pad_width, ndim, 'pad_width')
        
        constant_values_param = self._parse_param(data, self.constant_values, ndim, \
             'constant_values')

        end_values_param = self._parse_param(data, self.end_values, ndim, 'end_values')

        for dim in range(ndim):

            if (isinstance(data, ImageData)):
                if dimension_labels[dim] == 'channel':
                    geometry.channels += pad_width_param[dim][0]
                    geometry.channels += pad_width_param[dim][1]
                elif dimension_labels[dim] == 'vertical':
                    geometry.voxel_num_z += pad_width_param[dim][0]
                    geometry.voxel_num_z += pad_width_param[dim][1]
                elif dimension_labels[dim] == 'horizontal_x':
                    geometry.voxel_num_x += pad_width_param[dim][0]
                    geometry.voxel_num_x += pad_width_param[dim][1]
                elif dimension_labels[dim] == 'horizontal_y':
                    geometry.voxel_num_y += pad_width_param[dim][0]
                    geometry.voxel_num_y += pad_width_param[dim][1]
            
            # if AcquisitionData
            else:
                if dimension_labels[dim] == 'channel':
                    geometry.set_channels(num_channels=geometry_0.config.channels.num_channels + \
                        pad_width_param[dim][0] + pad_width_param[dim][1])
                elif dimension_labels[dim] == 'angle':
                    # pad angles vector
                    pad_width_param[dim] = (0,0)
                elif dimension_labels[dim] == 'vertical':
                    geometry.config.panel.num_pixels[1] += pad_width_param[dim][0]
                    geometry.config.panel.num_pixels[1] += pad_width_param[dim][1]
                elif dimension_labels[dim] == 'horizontal':
                    geometry.config.panel.num_pixels[0] += pad_width_param[dim][0]
                    geometry.config.panel.num_pixels[0] += pad_width_param[dim][1]
        
        if out == None:
            data_padded = geometry.allocate()
        else:
            if out.geometry != geometry:
                raise ValueError('The geometry in the argument out we received is not consistent with the requested padding.')
            data_padded = out
        
        
        if self.mode in ['reflect', 'symmetric', 'wrap', 'edge']:
            data_padded.fill(numpy.pad(data.as_array(), pad_width_param, mode=self.mode))
        elif self.mode == 'constant':
            data_padded.fill(numpy.pad(data.as_array(), pad_width_param, mode=self.mode, \
                constant_values=constant_values_param))
        elif self.mode == 'linear_ramp':
            data_padded.fill(numpy.pad(data.as_array(), pad_width_param, mode=self.mode, \
                end_values=end_values_param))
        
        if out == None:
            return data_padded


    def _parse_param(self, data, param, ndim, descr):

        if isinstance(param, Number):
            pad_param = [(int(param), int(param))] * ndim
        elif isinstance(param, tuple) and len(param) == 2:
            # create a list of tuples containing 2 ints in each tuple
            pad_param = [ tuple([int(el) for el in param]) ] * ndim
        elif isinstance(param, dict):
            pad_param = [(0,0)] * ndim
            for key in param.keys():
                if key == 'angle':
                    raise NotImplementedError('Cannot use Padder to pad the angle dimension')
                idx = data.dimension_labels.index(key)
                if isinstance(param[key], Number):
                    pad_param[idx] = (int(param[key]), int(param[key]))
                elif isinstance(param[key], tuple) and len(param[key]) == 2:
                    # create a tuple containing 2 ints
                    pad_param[idx] = tuple([int(el) for el in param[key]])
                else:
                    raise ValueError('Cannot parse provided {}. Expecting a number or tuple of length 2. Got {}'\
                        .format(descr, param[key]))
        else:
            raise ValueError('Cannot parse provided {}. Expecting int, tuple or dictionary with dimension lables. Got'\
                .format(descr, type(param)))

        return pad_param




# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np

from cil.optimisation.operators import DiagonalOperator

class MaskOperator(DiagonalOperator):
    
    r""" MaskOperator
   
    Parameters
    ----------
    mask : DataContainer
        Boolean array with the same dimensions as the data to be operated on.
    domain_geometry : ImageGeometry
        Specifies the geometry of the operator domain. If 'None' will use the mask geometry size and spacing as float32. default = None .
    """
    
    def __init__(self, mask, domain_geometry=None):

        #if domain_geometry is not specified assume float32 for domain_geometry data type
        if domain_geometry is None:
            domain_geometry = mask.geometry.copy()
            domain_geometry.dtype = np.float32

        super(MaskOperator, self).__init__(mask, domain_geometry)
        self.mask = self.diagonal
        
        




from cil.framework import AcquisitionGeometry
from cil.framework import cilacc
import numpy as np
import os
import logging

#set output level for all tests
logging.basicConfig(level=logging.WARNING)

#import cvxpy
try:
    import cvxpy
    has_cvxpy = True
except ModuleNotFoundError:
    has_cvxpy = False
print ("has_cvxpy\t{}".format(has_cvxpy))    

try:
    import tigre
    has_tigre = True
except ModuleNotFoundError:
    has_tigre = False
print ("has_tigre\t{}".format(has_tigre))


try:
    import astra
    has_astra = True
except ModuleNotFoundError:
    has_astra = False
print ("has_astra\t{}".format(has_astra))


try:
    cilacc.filter_projections_avh
    has_ipp = True
except:
    has_ipp = False
print ("has_ipp\t{}".format(has_ipp))


def has_nvidia_smi():
    return os.system('nvidia-smi') == 0

def has_gpu_tigre():

    if not has_tigre:
        return False

    has_gpu = True
    if has_nvidia_smi():
        from cil.plugins.tigre import ProjectionOperator
        from tigre.utilities.errors import TigreCudaCallError

        N = 3
        angles = np.linspace(0, np.pi, 2, dtype='float32')

        ag = AcquisitionGeometry.create_Cone2D([0,-100],[0,200])\
                                .set_angles(angles, angle_unit='radian')\
                                .set_panel(N, 0.1)\
                                .set_labels(['angle', 'horizontal'])
        
        ig = ag.get_ImageGeometry()
        
        data = ig.allocate(1)

        Op = ProjectionOperator(ig, ag)

        try:
            Op.direct(data)
            has_gpu = True
        except TigreCudaCallError:
            has_gpu = False
    else:
        has_gpu = False

    
    print ("has_gpu_tigre\t{}".format(has_gpu))
    return has_gpu


def has_gpu_astra():

    if not has_astra:
        return False

    has_gpu = False
    if has_nvidia_smi():
        try:
            astra.test_CUDA()
            has_gpu = True
        except:
            pass

    print ("has_gpu_astra\t{}".format(has_gpu))
    return has_gpu





# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import os
from cil.framework import AcquisitionData, AcquisitionGeometry, ImageData, ImageGeometry
from cil.version import version
import datetime

h5pyAvailable = True
try:
    import h5py 
except:
    h5pyAvailable = False


class NEXUSDataWriter(object):
    
    def __init__(self,
                 **kwargs):

        '''
        Constructor 

        :param data: The dataset to write to file
        :type data: AcquisitionData, ImageData
        :param file_name: file name to write
        :type file_name: os.path or string, default None
        :param compression: The lossy compression to apply, default 0 will not compress data. 8 or 16 will compress to 8 and 16 bit dtypes respectively.
        :type compression: int, default 0
        '''

        self.data = kwargs.get('data', None)
        self.file_name = kwargs.get('file_name', None)
        self.compression = kwargs.get('compression', 0)

        if ((self.data is not None) and (self.file_name is not None)):
            self.set_up(data = self.data,
                        file_name = self.file_name, compression=self.compression)
        
    def set_up(self,
               data = None,
               file_name = None,
               compression = 0):

        '''
        set up writer

        :param data: The dataset to write to file
        :type data: AcquisitionData, ImageData
        :param file_name: file name to write
        :type file_name: os.path or string, default None
        :param compression: The lossy compression to apply, default 0 will not compress data. 8 or 16 will compress to 8 and 16bit dtypes respectively.
        :type compression: int, default 0
        '''        
        self.data = data

        if not file_name.endswith('nxs') and not file_name.endswith('nex'):
            file_name+='.nxs'

        self.file_name = os.path.abspath(file_name)
        self.compression = compression
        
        if not ((isinstance(self.data, ImageData)) or 
                (isinstance(self.data, AcquisitionData))):
            raise Exception('Writer supports only following data types:\n' +
                            ' - ImageData\n - AcquisitionData')

        if self.compression == 0:
            self.dtype = data.dtype
            self.compress = False
        elif self.compression == 8:
            self.dtype = np.uint8
            self.compress = True
        elif self.compression == 16:
            self.dtype = np.uint16
            self.compress = True
        else:
            raise Exception('Compression bits not valid. Got {0} expected value in {1}'.format(self.compression, [0,8,16]))

        # check that h5py library is installed
        if (h5pyAvailable == False):
            raise Exception('h5py is not available, cannot write NEXUS files.')
    
    def write(self):

        '''
        write dataset to disk
        '''   
        # if the folder does not exist, create the folder
        if not os.path.isdir(os.path.dirname(self.file_name)):
            os.mkdir(os.path.dirname(self.file_name))

        if self.compress is True:
            save_range = np.iinfo(self.dtype).max

            data_min = self.data.min()
            data_range = self.data.max() - data_min

            if data_range > 0:
                scale = save_range / data_range
                offset = - data_min * scale
            else:
                scale = 1.0
                offset = 0.0
                
        # create the file
        with h5py.File(self.file_name, 'w') as f:
            
            # give the file some important attributes
            f.attrs['file_name'] = self.file_name
            f.attrs['cil_version'] = version
            f.attrs['file_time'] = str(datetime.datetime.utcnow())
            f.attrs['creator'] = np.string_('NEXUSDataWriter.py')
            f.attrs['NeXus_version'] = '4.3.0'
            f.attrs['HDF5_Version'] = h5py.version.hdf5_version
            f.attrs['h5py_version'] = h5py.version.version
            
            # create the NXentry group
            nxentry = f.create_group('entry1/tomo_entry')
            nxentry.attrs['NX_class'] = 'NXentry'

            #create empty data entry
            ds_data = f.create_dataset('entry1/tomo_entry/data/data',shape=self.data.shape, dtype=self.dtype)

            if self.compress:
                ds_data.attrs['scale'] = scale
                ds_data.attrs['offset'] = offset

                for i in range(self.data.shape[0]):
                    ds_data[i:(i+1)] = self.data.array[i] * scale + offset
            else:
                ds_data.write_direct(self.data.array)

            # set up dataset attributes
            if (isinstance(self.data, ImageData)):
                ds_data.attrs['data_type'] = 'ImageData'
            else:
                ds_data.attrs['data_type'] = 'AcquisitionData'
            
            for i in range(self.data.number_of_dimensions):
                ds_data.attrs['dim{}'.format(i)] = self.data.dimension_labels[i]
            
            if (isinstance(self.data, AcquisitionData)): 
                
                # create group to store configuration
                f.create_group('entry1/tomo_entry/config')
                f.create_group('entry1/tomo_entry/config/source')
                f.create_group('entry1/tomo_entry/config/detector')
                f.create_group('entry1/tomo_entry/config/rotation_axis')
                
                ds_data.attrs['geometry'] = self.data.geometry.config.system.geometry
                ds_data.attrs['dimension'] = self.data.geometry.config.system.dimension   
                ds_data.attrs['num_channels'] = self.data.geometry.config.channels.num_channels
                
                f.create_dataset('entry1/tomo_entry/config/detector/direction_x', 
                                 (self.data.geometry.config.system.detector.direction_x.shape), 
                                 dtype = 'float32', 
                                 data = self.data.geometry.config.system.detector.direction_x)
                
                f.create_dataset('entry1/tomo_entry/config/detector/position', 
                                 (self.data.geometry.config.system.detector.position.shape), 
                                 dtype = 'float32', 
                                 data = self.data.geometry.config.system.detector.position)
                
                if self.data.geometry.config.system.geometry == 'cone':
                    f.create_dataset('entry1/tomo_entry/config/source/position', 
                                     (self.data.geometry.config.system.source.position.shape), 
                                     dtype = 'float32', 
                                     data = self.data.geometry.config.system.source.position)
                else:
                    f.create_dataset('entry1/tomo_entry/config/ray/direction', 
                                     (self.data.geometry.config.system.ray.direction.shape), 
                                     dtype = 'float32', 
                                     data = self.data.geometry.config.system.ray.direction)
                
                f.create_dataset('entry1/tomo_entry/config/rotation_axis/position', 
                                 (self.data.geometry.config.system.rotation_axis.position.shape), 
                                 dtype = 'float32', 
                                 data = self.data.geometry.config.system.rotation_axis.position)
                
                
                ds_data.attrs['num_pixels_h'] = self.data.geometry.config.panel.num_pixels[0]
                ds_data.attrs['pixel_size_h'] = self.data.geometry.config.panel.pixel_size[0]
                ds_data.attrs['panel_origin'] = self.data.geometry.config.panel.origin

                if self.data.geometry.config.system.dimension == '3D':
                    f.create_dataset('entry1/tomo_entry/config/detector/direction_y', 
                                     (self.data.geometry.config.system.detector.direction_y.shape), 
                                     dtype = 'float32', 
                                     data = self.data.geometry.config.system.detector.direction_y)
                    
                    f.create_dataset('entry1/tomo_entry/config/rotation_axis/direction', 
                                 (self.data.geometry.config.system.rotation_axis.direction.shape), 
                                 dtype = 'float32', 
                                 data = self.data.geometry.config.system.rotation_axis.direction)
                
                    ds_data.attrs['num_pixels_v'] = self.data.geometry.config.panel.num_pixels[1]
                    ds_data.attrs['pixel_size_v'] = self.data.geometry.config.panel.pixel_size[1]
                    
                angles = f.create_dataset('entry1/tomo_entry/config/angles', 
                                                 (self.data.geometry.config.angles.angle_data.shape), 
                                                 dtype = 'float32', 
                                                 data = self.data.geometry.config.angles.angle_data)
                angles.attrs['angle_unit'] = self.data.geometry.config.angles.angle_unit   
                angles.attrs['initial_angle'] = self.data.geometry.config.angles.initial_angle
            
            else:   # ImageData
                
                ds_data.attrs['voxel_num_x'] = self.data.geometry.voxel_num_x
                ds_data.attrs['voxel_num_y'] = self.data.geometry.voxel_num_y
                ds_data.attrs['voxel_num_z'] = self.data.geometry.voxel_num_z
                ds_data.attrs['voxel_size_x'] = self.data.geometry.voxel_size_x
                ds_data.attrs['voxel_size_y'] = self.data.geometry.voxel_size_y
                ds_data.attrs['voxel_size_z'] = self.data.geometry.voxel_size_z
                ds_data.attrs['center_x'] = self.data.geometry.center_x
                ds_data.attrs['center_y'] = self.data.geometry.center_y
                ds_data.attrs['center_z'] = self.data.geometry.center_z
                ds_data.attrs['channels'] = self.data.geometry.channels
                ds_data.attrs['channel_spacing'] = self.data.geometry.channel_spacing



# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from cil.optimisation.functions import Function, IndicatorBox
from cil.optimisation.operators import GradientOperator
import numpy as np
from numbers import Number
import warnings



class TotalVariation(Function):
    
    r'''Fast Gradient Projection algorithm for Total Variation(TV) regularisation
    
    .. math::  \min_{x} \frac{1}{2}||x-b||^{2}_{2} + \tau\alpha TV(x)
                
    Parameters:
      
      :param max_iteration: max iterations of FGP algorithm
      :type max_iteration: int, default 100
      :param tolerance: Stopping criterion
      :type tolerance: float, default `None` 
      :param correlation: Correlation between `Space` and/or `SpaceChannels` for the GradientOperator
      :type correlation: str, default 'Space'
      :param backend: Backend to compute finite differences for the GradientOperator
      :type backend: str, default 'c'
      :param lower: lower bound for the orthogonal projection onto the convex set C
      :type lower: Number, default `-np.inf`
      :param upper: upper bound for the orthogonal projection onto the convex set C
      :type upper: Number, default `+np.inf`
      :param isotropic: L2 norm is used for Gradient Operator (isotropic) 
      :type isotropic: bool, default `True` 
      
                        .. math:: \sum \sqrt{(\partial_y u)^{2} + (\partial_x u)^2} \mbox{ (isotropic) }

                        .. math:: \sum |\partial_y u| + |\partial_x u| \mbox{ (anisotropic) }
       
      :param split: splits the Gradient into spatial Gradient and spectral Gradient for multichannel data
      :type split: bool, default `False`           
      :param info: force a print to screen stating the stop
      :type info: bool, default `False`

      :Example:
 
      TV = alpha * TotalVariation()
      sol = TV.proximal(data, tau = 1.0) 

      .. note:: `tau` can be a number or an array. The latter case implies that step-size preconditioning is applied.

    Reference:
      
        A. Beck and M. Teboulle, "Fast Gradient-Based Algorithms for Constrained Total Variation 
        Image Denoising and Deblurring Problems," in IEEE Transactions on Image Processing,
        vol. 18, no. 11, pp. 2419-2434, Nov. 2009, 
        doi: 10.1109/TIP.2009.2028250.
        
    '''    
    
    
    def __init__(self,
                 max_iteration=100, 
                 tolerance = None, 
                 correlation = "Space",
                 backend = "c",
                 lower = -np.inf, 
                 upper = np.inf,
                 isotropic = True,
                 split = False,
                 info = False):
        

        super(TotalVariation, self).__init__(L = None)
        # Regularising parameter = alpha
        self.regularisation_parameter = 1.
        
        # Iterations for FGP_TV
        self.iterations = max_iteration
        
        # Tolerance for FGP_TV
        self.tolerance = tolerance
        
        # Total variation correlation (isotropic=Default)
        self.isotropic = isotropic
        
        # correlation space or spacechannels
        self.correlation = correlation
        self.backend = backend        
        
        # Define orthogonal projection onto the convex set C
        self.lower = lower
        self.upper = upper
        self.tmp_proj_C = IndicatorBox(lower, upper).proximal
                        
        # Setup GradientOperator as None. This is to avoid domain argument in the __init__     

        self._gradient = None
        self._domain = None

        self.pptmp = None
        self.pptmp1 = None
        
        # Print stopping information (iterations and tolerance error) of FGP_TV  
        self.info = info

        # splitting Gradient
        self.split = split

    @property
    def regularisation_parameter(self):
        return self._regularisation_parameter

    @regularisation_parameter.setter
    def regularisation_parameter(self, value):
        if not isinstance(value, Number):
            raise TypeError("regularisation_parameter: expectec a number, got {}".format(type(value)))
        self._regularisation_parameter = value

    def __call__(self, x):
        
        r''' Returns the value of the \alpha * TV(x)'''
        try:
            self._domain = x.geometry
        except:
            self._domain = x
        # evaluate objective function of TV gradient
        if self.isotropic:
            return self.regularisation_parameter * self.gradient.direct(x).pnorm(2).sum()
        else:
            return self.regularisation_parameter * self.gradient.direct(x).pnorm(1).sum() 
    
    
    def projection_C(self, x, out=None):   
                     
        r''' Returns orthogonal projection onto the convex set C'''

        try:
            self._domain = x.geometry
        except:
            self._domain = x
        return self.tmp_proj_C(x, tau = None, out = out)
                        
    def projection_P(self, x, out=None):
                       
        r''' Returns the projection P onto \|\cdot\|_{\infty} '''  
        try:
            self._domain = x.geometry
        except:
            self._domain = x
        
        # preallocated in proximal
        tmp = self.pptmp
        tmp1 = self.pptmp1
        tmp1 *= 0
        
        if self.isotropic:
            for el in x.containers:
                el.conjugate().multiply(el, out=tmp)
                tmp1.add(tmp, out=tmp1)
            tmp1.sqrt(out=tmp1)
            tmp1.maximum(1.0, out=tmp1)
            if out is None:
                return x.divide(tmp1)
            else:
                x.divide(tmp1, out=out)
        else:
            tmp1 = x.abs()
            tmp1.maximum(1.0, out=tmp1) 
            if out is None:
                return x.divide(tmp1)
            else:
                x.divide(tmp1, out=out)                   
    
    
    def proximal(self, x, tau, out = None):
        
        ''' Returns the solution of the FGP_TV algorithm '''         
        try:
            self._domain = x.geometry
        except:
            self._domain = x
        
        # initialise
        t = 1        
        tmp_p = self.gradient.range_geometry().allocate(0)  
        tmp_q = tmp_p.copy()
        tmp_x = self.gradient.domain_geometry().allocate(0)     
        p1 = self.gradient.range_geometry().allocate(0)
        

        should_break = False
        for k in range(self.iterations):
                                                                                   
            t0 = t
            self.gradient.adjoint(tmp_q, out = tmp_x)
            
            # axpby now works for matrices
            tmp_x.axpby(-self.regularisation_parameter*tau, 1.0, x, out=tmp_x)
            self.projection_C(tmp_x, out = tmp_x)                       

            self.gradient.direct(tmp_x, out=p1)
            if isinstance (tau, (Number, np.float32, np.float64)):
                p1 *= self.L/(self.regularisation_parameter * tau)
            else:
                p1 *= self.L/self.regularisation_parameter
                p1 /= tau

            if self.tolerance is not None:
                
                if k%5==0:
                    error = p1.norm()
                    p1 += tmp_q
                    error /= p1.norm()
                    if error<=self.tolerance:                           
                        should_break = True
                else:
                    p1 += tmp_q
            else:
                p1 += tmp_q
            if k == 0:
                # preallocate for projection_P
                self.pptmp = p1.get_item(0) * 0
                self.pptmp1 = self.pptmp.copy()

            self.projection_P(p1, out=p1)
            

            t = (1 + np.sqrt(1 + 4 * t0 ** 2)) / 2
            
            #tmp_q.fill(p1 + (t0 - 1) / t * (p1 - tmp_p))
            p1.subtract(tmp_p, out=tmp_q)
            tmp_q *= (t0-1)/t
            tmp_q += p1
            
            tmp_p.fill(p1)

            if should_break:
                break
        
        #clear preallocated projection_P arrays
        self.pptmp = None
        self.pptmp1 = None
        
        # Print stopping information (iterations and tolerance error) of FGP_TV     
        if self.info:
            if self.tolerance is not None:
                print("Stop at {} iterations with tolerance {} .".format(k, error))
            else:
                print("Stop at {} iterations.".format(k))                
            
        if out is None:                        
            self.gradient.adjoint(tmp_q, out=tmp_x)
            tmp_x *= tau
            tmp_x *= self.regularisation_parameter 
            x.subtract(tmp_x, out=tmp_x)
            return self.projection_C(tmp_x)
        else:          
            self.gradient.adjoint(tmp_q, out = out)
            out*=tau
            out*=self.regularisation_parameter
            x.subtract(out, out=out)
            self.projection_C(out, out=out)
    
    def convex_conjugate(self,x):        
        return 0.0    
    @property
    def L(self):
        if self._L is None:
            self.calculate_Lipschitz()
        return self._L
    @L.setter
    def L(self, value):
        warnings.warn("You should set the Lipschitz constant with calculate_Lipschitz().")
        if isinstance(value, (Number,)) and value >= 0:
            self._L = value
        else:
            raise TypeError('The Lipschitz constant is a real positive number')

    def calculate_Lipschitz(self):
        # Compute the Lipschitz parameter from the operator if possible
        # Leave it initialised to None otherwise
        self._L = (1./self.gradient.norm())**2  
    
    @property
    def gradient(self):
        '''creates a gradient operator if not instantiated yet

        There is no check that the variable _domain is changed after instantiation (should not be the case)'''
        if self._gradient is None:
            if self._domain is not None:
                self._gradient = GradientOperator(self._domain, correlation = self.correlation, backend = self.backend)
        return self._gradient
    def __rmul__(self, scalar):
        if not isinstance (scalar, Number):
            raise TypeError("scalar: Expectec a number, got {}".format(type(scalar)))
        self.regularisation_parameter = scalar
        return self

    

